//       _________ __                 __
//      /   _____//  |_____________ _/  |______     ____  __ __  ______
//      \_____  \\   __\_  __ \__  \\   __\__  \   / ___\|  |  \/  ___/
//      /        \|  |  |  | \// __ \|  |  / __ \_/ /_/  >  |  /\___ |
//     /_______  /|__|  |__|  (____  /__| (____  /\___  /|____//____  >
//             \/                  \/          \//_____/            \/
//  ______________________                           ______________________
//                        T H E   W A R   B E G I N S
//         Stratagus - A free fantasy real time strategy game engine
//
//      (c) Copyright 2000-2021 by Vladi Belperchinov-Shabanski and Andrettin
//
//      This program is free software; you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation; only version 2 of the License.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program; if not, write to the Free Software
//      Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
//      02111-1307, USA.

#pragma once

class CConfigData;
class CPlayer;
class CUnit;
class CUpgrade;

namespace wyrmgus {

class button;
class season;
class site;
class sml_data;
class sml_property;
class trigger;
class unit_type;

class condition
{
public:
	static std::unique_ptr<const condition> from_sml_property(const sml_property &property);
	static std::unique_ptr<const condition> from_sml_scope(const sml_data &scope);

	static std::string get_conditions_string(const std::vector<std::unique_ptr<const condition>> &conditions, const size_t indent)
	{
		std::string conditions_string;
		bool first = true;
		for (const std::unique_ptr<const condition> &condition : conditions) {
			if (condition->is_hidden()) {
				continue;
			}

			const std::string condition_string = condition->get_string(indent);
			if (condition_string.empty()) {
				continue;
			}

			if (first) {
				first = false;
			} else {
				conditions_string += "\n";
			}

			if (indent > 0) {
				conditions_string += std::string(indent, '\t');
			}

			conditions_string += condition_string;
		}
		return conditions_string;
	}

	virtual ~condition() {}

	void ProcessConfigData(const CConfigData *config_data);
	virtual void ProcessConfigDataProperty(const std::pair<std::string, std::string> &property);
	virtual void ProcessConfigDataSection(const CConfigData *section);
	virtual void process_sml_property(const sml_property &property);
	virtual void process_sml_scope(const sml_data &scope);

	virtual void check_validity() const
	{
	}

	virtual bool check(const CPlayer *player, bool ignore_units = false) const = 0;
	virtual bool check(const CUnit *unit, bool ignore_units = false) const;

	//get the condition as a string
	virtual std::string get_string(const size_t indent) const = 0;

	virtual bool is_hidden() const
	{
		return false;
	}
};

template <bool precondition>
extern bool check_special_conditions(const unit_type *target, const CPlayer *player, const bool ignore_units, const bool is_neutral_use);

template <bool precondition>
extern bool check_special_conditions(const CUpgrade *target, const CPlayer *player, const bool ignore_units, const bool is_neutral_use);

//check conditions for player
template <bool precondition = false, typename T>
extern bool check_conditions(const T *target, const CPlayer *player, const bool ignore_units = false, const bool is_neutral_use = false)
{
	if constexpr (!precondition) {
		if (!check_conditions<true>(target, player, ignore_units, is_neutral_use)) {
			return false;
		}
	}

	if constexpr (std::is_same_v<T, unit_type>) {
		if (!check_special_conditions<precondition>(target, player, ignore_units, is_neutral_use)) {
			return false;
		}
	} else if constexpr (std::is_same_v<T, CUpgrade>) {
		if (!check_special_conditions<precondition>(target, player, ignore_units, is_neutral_use)) {
			return false;
		}
	}

	if constexpr (precondition) {
		return target->get_preconditions() == nullptr || target->get_preconditions()->check(player, ignore_units);
	} else {
		return target->get_conditions() == nullptr || target->get_conditions()->check(player, ignore_units);
	}
}

template <bool precondition>
extern bool check_special_conditions(const unit_type *target, const CUnit *unit, const bool ignore_units);

template <bool precondition>
extern bool check_special_conditions(const CUpgrade *target, const CUnit *unit, const bool ignore_units);

//check conditions for unit
template <bool precondition = false, typename T>
extern bool check_conditions(const T *target, const CUnit *unit, const bool ignore_units = false)
{
	if constexpr (!precondition) {
		if (!check_conditions<true>(target, unit, ignore_units)) {
			return false;
		}
	}

	if constexpr (std::is_same_v<T, unit_type>) {
		if (!check_special_conditions<precondition>(target, unit, ignore_units)) {
			return false;
		}
	} else if constexpr (std::is_same_v<T, CUpgrade>) {
		if (!check_special_conditions<precondition>(target, unit, ignore_units)) {
			return false;
		}
	}

	if constexpr (precondition) {
		return target->get_preconditions() == nullptr || target->get_preconditions()->check(unit, ignore_units);
	} else {
		return target->get_conditions() == nullptr || target->get_conditions()->check(unit, ignore_units);
	}
}

}

/// Register CCL features for dependencies
extern void DependenciesCclRegister();

/// Print all unit conditions into string
extern std::string PrintConditions(const wyrmgus::button &button);
